/**
 * @returns {Modeling}
 */
function getCurrentModeler() {
  return dmnModeler.getActiveViewer().get("modeling");
}

function getShapeWithId(id) {
  return getCurrentModeler()._canvas._elementRegistry.get(id);
}

function getSheetElementWithId(id) {
  return getCurrentModeler()._sheet._elementRegistry.get(id);
}

/**
 *
 * @param sourceX
 * @param sourceY
 * @param w
 * @param offsetX
 * @param offsetY
 * @returns {{x: *, y: *}}
 */
function getRandomNewPointInCycle(sourceX, sourceY, w, offsetX, offsetY) {
  let dx = Math.floor(Math.random() * Math.floor(2 * w)) - w;
  let dy = Math.sqrt(w * w - dx * dx);
  if (Math.random() > 0.5) {
    dy *= -1;
  }
  if (offsetX) {
    dx += offsetX;
  }
  if (offsetY) {
    dy += offsetY;
  }
  return {x: sourceX + dx, y: sourceY + dy};
}

/**
 *
 * @param {VerificationEntry}  verificationEntry
 * @param {VerificationFix} fix
 * @param id fix row id
 * @param $callerButton
 */
function performVerificationFix(verificationEntry, fix, id, $callerButton) {
  console.log("FIX", fix);
  // highlight row
  cleanHighlightFunction();
  $('.verification-container').removeClass('highlight-soft');
  $('#ver-entry-' + id).addClass('highlight-soft');
  let cud = false;
  for (let i = 0; i < fix.actions.length; i++) {
    switch (fix.actions[i].actionType) {
      case 'SHOW':
        performVerificationFixSHOW(verificationEntry, fix.actions[i]);
        break;
      case 'CREATE':
        performVerificationFixCREATE(verificationEntry, fix.actions[i]);
        cud = true;
        break;
      case 'DELETE':
        performVerificationFixDELETE(verificationEntry, fix.actions[i]);
        cud = true;
        break;
      case 'UPDATE':
        performVerificationFixUPDATE(verificationEntry, fix.actions[i]);
        cud = true;
        break;
      default:
        alert("Action not defined: " + fix);
    }
    if (cud) {
      $callerButton.css('display', 'none');
      if ($checkBoxReverify[0].checked) {
        checkVerifications();
      }
    }
    performHighlightFunction();
  }
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function performVerificationFixSHOW(verificationEntry, fixAction) {
  switch (fixAction.actionScope) {
    case 'RULE':
      fixSHOW_RULE(verificationEntry, fixAction);
      break;
    case 'INPUT_ENTRY':
      fixSHOW_INPUT_ENTRY(verificationEntry, fixAction);
      break;
    case 'OUTPUT_ENTRY':
      fixSHOW_OUTPUT_ENTRY(verificationEntry, fixAction);
      break;
    case 'INPUT_DATA':
      fixSHOW_INPUT_DATA(verificationEntry, fixAction);
      break;
    case 'INPUT_COLUMN':
      fixSHOW_INPUT_COLUMN(verificationEntry, fixAction);
      break;
    case 'OUTPUT_COLUMN':
      fixSHOW_OUTPUT_COLUMN(verificationEntry, fixAction);
      break;
    case 'DECISION_TABLE':
      fixSHOW_DECISION_TABLE(verificationEntry, fixAction);
      break;
    case 'DECISION':
      fixSHOW_DECISION(verificationEntry, fixAction);
      break;
    default:
      alert("ACTION: " + fix.actions[i].actionType + ' -> '
          + fix.actions[i].actionScope); // TODO
  }
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function fixSHOW_RULE(verificationEntry, fixAction) {
  highlightRules(verificationEntry.elements);
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function fixSHOW_INPUT_ENTRY(verificationEntry, fixAction) {
  highlightDataEntries(verificationEntry.elements, 'inputEntryId');
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function fixSHOW_INPUT_DATA(verificationEntry, fixAction) {
  highlightInputData(verificationEntry.elements);
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function fixSHOW_OUTPUT_ENTRY(verificationEntry, fixAction) {
  highlightDataEntries(verificationEntry.elements, 'outputEntryId');
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function fixSHOW_INPUT_COLUMN(verificationEntry, fixAction) {
  highlightColumns(verificationEntry.elements, 'inputId');
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function fixSHOW_OUTPUT_COLUMN(verificationEntry, fixAction) {
  highlightColumns(verificationEntry.elements, 'outputId');
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function fixSHOW_DECISION_TABLE(verificationEntry, fixAction) {
  highlightDecision(verificationEntry.elements);
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function fixSHOW_DECISION(verificationEntry, fixAction) {
  highlightDecision(verificationEntry.elements);
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function performVerificationFixCREATE(verificationEntry, fixAction) {
  switch (fixAction.actionScope) {
    case 'RULE':
      createRule(verificationEntry, fixAction);
      break;
    case 'INPUT_DATA':
      createInputData(verificationEntry, fixAction);
      break;
    case 'INPUT_COLUMN':
      createInputColumn(verificationEntry, fixAction);
      break;
    case 'INPUT_ENTRY':
    case 'OUTPUT_ENTRY':
    case 'OUTPUT_COLUMN':
    case 'DECISION_TABLE':
    case 'DECISION':
    default:
      alert("ACTION undefined: " + fixAction.actionType + ' -> '
          + fixAction.actionScope);
  }
}

/**
 *
 * @param {VerificationEntry}  verificationEntry
 * @param {Action} fixAction
 */
function createRule(verificationEntry, fixAction) {
  openViewWithId(fixAction.actionValues['decisionId']);
  const modeler = getCurrentModeler();
  const rule = modeler.addRow({type: "dmn:DecisionRule"});
  const {cells} = rule;
  for (const [key, value] of Object.entries(fixAction.actionValues)) {
    let cell = getCellByColId(key, cells);
    if (cell) {
      modeler.editCell(getCellByColId(key, cells), value);
    }
  }
}

/**
 *
 * @param {VerificationEntry}  verificationEntry
 * @param {Action} fixAction
 */
function createInputData(verificationEntry, fixAction) {
  openViewWithId(verificationEntry.elements[0].identifier['definitionId']);
  let decision = getShapeWithId(fixAction.actionValues['decisionId']);
  const modeler = getCurrentModeler();
  const shape = modeler._elementFactory.createDrdElement('shape',
      {type: 'dmn:InputData'});
  let inputData = modeler.appendShape(decision, shape,
      getRandomNewPointInCycle(decision.x, decision.y, 250, decision.width / 2,
          decision.height / 2),
      getShapeWithId(verificationEntry.elements[0].identifier['definitionId']),
      {attach: false, connection: {type: "dmn:InformationRequirement"}}
  );
  inputData.businessObject.name = fixAction.actionValues['name'];
  openViewWithId(verificationEntry.elements[0].identifier['definitionId']);
}

/**
 *
 * @param {VerificationEntry}  verificationEntry
 * @param {Action} fixAction
 */
function createInputColumn(verificationEntry, fixAction) {
  openViewWithId(fixAction.actionValues['decisionId']);
  const modeler = getCurrentModeler();
  let col = modeler.addCol({type: 'dmn:InputClause'});
  col.businessObject.inputExpression.typeRef = fixAction.actionValues['typeRef'].toLowerCase();
  col.businessObject.inputExpression.text = fixAction.actionValues['name'];
  openViewWithId(fixAction.actionValues['decisionId']);
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function performVerificationFixDELETE(verificationEntry, fixAction) {
  switch (fixAction.actionScope) {
    case 'RULE':
      deleteRule(verificationEntry, fixAction);
      break;
    case 'INPUT_COLUMN':
      deleteColumn(verificationEntry, fixAction, 'inputId');
      break;
    case 'OUTPUT_COLUMN':
      deleteColumn(verificationEntry, fixAction, 'outputId');
      break;
    case 'INPUT_DATA':
      deleteInputData(verificationEntry, fixAction);
      break;
    case 'INPUT_ENTRY':
    case 'OUTPUT_ENTRY':
    case 'DECISION_TABLE':
    case 'DECISION':
    default:
      alert("ACTION undefined: " + fixAction.actionType + ' -> '
          + fixAction.actionScope);
  }
}

/**
 *
 * @param {VerificationEntry}  verificationEntry
 * @param {Action} fixAction
 */
function deleteRule(verificationEntry, fixAction) {
  openViewWithId(fixAction.actionValues['decisionId']);
  const modeler = getCurrentModeler();
  const row = modeler._sheet._elementRegistry.get(
      fixAction.actionValues['ruleId']);
  if (row) {
    modeler.removeRow(row);
  }
}

/**
 *
 * @param {VerificationEntry}  verificationEntry
 * @param {Action} fixAction
 * @param {String} columnIdName
 */
function deleteColumn(verificationEntry, fixAction, columnIdName) {
  openViewWithId(fixAction.actionValues['decisionId']);
  const modeler = getCurrentModeler();
  const col = getSheetElementWithId(fixAction.actionValues[columnIdName]);
  if (col) {
    modeler.removeCol(col);
  }
}

/**
 *
 * @param {VerificationEntry}  verificationEntry
 * @param {Action} fixAction
 */
function deleteInputData(verificationEntry, fixAction) {
  openViewWithId(verificationEntry.elements[0].identifier['definitionId']);
  const modeler = getCurrentModeler();
  let inputData = getShapeWithId(fixAction.actionValues['inputDataId']);
  if (inputData) {
    modeler.removeShape(inputData);
  }
}

/**
 *
 * @param {VerificationEntry} verificationEntry
 * @param {Action} fixAction
 */
function performVerificationFixUPDATE(verificationEntry, fixAction) {
  switch (fixAction.actionScope) {
    case 'INPUT_DATA':
      updateInputData(verificationEntry, fixAction);
      break;
    case 'DECISION':
      updateDecision(verificationEntry, fixAction);
      break;
    case 'RULE':
    case 'INPUT_COLUMN':
    case 'OUTPUT_COLUMN':
    case 'INPUT_ENTRY':
    case 'OUTPUT_ENTRY':
    case 'DECISION_TABLE':
    default:
      alert("ACTION undefined: " + fixAction.actionType + ' -> '
          + fixAction.actionScope);
  }
}

/**
 *
 * @param {VerificationEntry}  verificationEntry
 * @param {Action} fixAction
 */
function updateInputData(verificationEntry, fixAction) {
  openViewWithId(verificationEntry.elements[0].identifier['definitionId']);
  let inputData = getShapeWithId(fixAction.actionValues['inputDataId']);
  if (fixAction.actionValues['decisionId']) {
    const modeler = getCurrentModeler();
    let decision = getShapeWithId(fixAction.actionValues['decisionId']);
    getCurrentModeler().connect(inputData, decision,
        {type: "dmn:InformationRequirement"})
  }
  if (fixAction.actionValues['name']) {
    inputData.businessObject.name = fixAction.actionValues['name'];
  }
  openViewWithId(verificationEntry.elements[0].identifier['definitionId']);
}

/**
 *
 * @param {VerificationEntry}  verificationEntry
 * @param {Action} fixAction
 */
function updateDecision(verificationEntry, fixAction) {
  openViewWithId(verificationEntry.elements[0].identifier['definitionId']);
  let decision = getShapeWithId(
      fixAction.actionValues['decisionId']);
  if (fixAction.actionValues['targetDecisionId']) {
    let targetDecision = getShapeWithId(
        fixAction.actionValues['targetDecisionId']);
    const modeler = getCurrentModeler();
    getCurrentModeler().connect(decision, targetDecision,
        {type: "dmn:InformationRequirement"})
  }
  if (fixAction.actionValues['name']) {
    decision.businessObject.name = fixAction.actionValues['name'];
  }
  openViewWithId(verificationEntry.elements[0].identifier['definitionId']);
}

// get a rule:
/**
 * Get a col by ID
 * @param colId
 * @param cells
 * @returns {undefined|*}
 */
function getCellByColId(colId, cells) {
  for (let i = 0; i < cells.length; i++) {
    if (cells[i].col.id === colId) {
      return cells[i];
    }
  }
  return undefined;
}
